#include "Esimple.h"
#include "hexmesh.h"
#include "q2h.h"
#include "split.h"
#include "qsplit.h"
#include "cracksurf.h"

extern int
generate_circle_quadrant(int nv_per_side, double radius);
extern int
generate_sphere_octant(int nv_per_side, double radius);

static hexmesh_model_t *hexmesh_model = NULL;
static char            *hexmesh_model_description = NULL;
static WCRec            conc_load_components = {0, 0, 0};
static int              constraint_type = 'X';
static double           material_E = 210000.0;
static double           material_nu = 0.3;
static double           material_rho = 7.85;
static int              curr_partition = 1;
static int              quadrature_order = 2;
static int              mat_id = 1;
static int              subsplit = 0;
static double           diag_length_mult = 1;
static int              symbolic_tractions = 0;
static int              nodes_at_hex_centers = 0;
static double           rand_pos_ampl = 0;
static double           search_tol = FLT_MAX;
static double           user_search_tol = 0;
static int              nx = 0, ny = 0, nz = 0;
static int              crack_surf_format = CRACK_SURF_FORMAT;

#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

static void 
do_make_hexmesh_model(EView *v_p, caddr_t data_p);
static void
print_hexmesh_model_CB(Widget w, XtPointer ptr, XtPointer call);
static void
print_tees_model_CB(Widget w, XtPointer ptr, XtPointer call);
static void
print_ndot_model_CB(Widget w, XtPointer ptr, XtPointer call);
static void
print_hexmesh_model_parallel_CB(Widget w, XtPointer ptr, XtPointer call);
static void 
set_conc_load_components(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
set_function(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
set_constraint_type(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
set_rand_pos_ampl(Widget w, XtPointer text_ptr, XtPointer call_data);
static void
make_constraint_CB(Widget w, XtPointer ptr, XtPointer call);
static void
rand_pos_CB(Widget w, XtPointer ptr, XtPointer call);
static void
make_conc_load_CB(Widget w, XtPointer ptr, XtPointer call);
static void
make_surf_tractions_CB(Widget w, XtPointer ptr, XtPointer call);
static void 
do_make_constraint(EView *v_p, caddr_t data_p);
static void 
do_rand_pos(EView *v_p, caddr_t data_p);
static void 
do_make_conc_load(EView *v_p, caddr_t data_p);
static void 
do_make_surf_tractions(EView *v_p, caddr_t data_p);
static void 
set_hexmesh_model_description(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
change_to_efg_CB(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
change_quadrature_CB(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
change_subsplit_CB(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
compute_d_ms_CB(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
do_change_to_efg(EView *v_p, caddr_t data_p);
static void 
do_change_quadrature(EView *v_p, caddr_t data_p);
static void 
set_E_nu(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
set_diag_length_mult(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
set_search_tol(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
set_quadrature_order(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
set_subsplit(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
set_partition(Widget w, XtPointer text_ptr, XtPointer call_data);
static void
make_output_request_CB(Widget w, XtPointer ptr, XtPointer call);
static void
make_output_request_by_loc_CB(Widget w, XtPointer ptr, XtPointer call);
static void 
do_make_output_request_by_loc(EView *v_p, caddr_t data_p, WCRec *p);
static void 
do_make_output_request(EView *v_p, caddr_t data_p);
static void 
add_to_partition_CB(Widget w, XtPointer text_ptr, XtPointer call_data);
static void
set_extrusion_vector(Widget w, XtPointer client_data, XtPointer call_data);
static void
extrude_quad_to_hex_CB(Widget w, XtPointer client_data, XtPointer call_data);
static void
split_hex_CB(Widget w, XtPointer client_data, XtPointer call_data);
static void
split_quad_CB(Widget w, XtPointer client_data, XtPointer call_data);
static void
set_split_intervals(Widget w, XtPointer client_data, XtPointer call_data);
static void
load_and_eval_CB(Widget w, XtPointer client_data, XtPointer call_data);
static void
toggle_symbolic_tracs_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
toggle_nodes_at_hex_centers_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
randomize_diagonals_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
construct_crack_surf_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
construct_smf_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
print_wasp_trac_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
draw_circle_quadrant_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
draw_sphere_octant_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
set_radius_and_nv(Widget w, XtPointer client_data, XtPointer call_data);
static void
count_hexes_CB(Widget w, XtPointer ptr, XtPointer call_data);



static void
make_hexmesh_model_CB(Widget w, XtPointer ptr, XtPointer call)
{
  EVSetSelectActionFunction(do_make_hexmesh_model, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ESIHandleCmd ("COLOR RED");
  ESIHandleCmd ("LAYER ACTIVE 0");
  ESIHandleCmd ("LAYER OFF 1313");
  ESIHandleCmd ("MTYPE 1");
  ESIHandleCmd ("MSIZE 6");
  ERptStatusMessage(1, 1, "Select hexahedrons to be included in the mesh");
}


static void 
do_make_hexmesh_model(EView *v_p, caddr_t data_p)
{
  GraphicObj *hex = NULL;
  int i;
  WCRec pts[8], v;
  int np;
  node_t *nodes[8];
  double d, sumd;
  int ch, nh;
  
  if ((nh = EVItemsInSelectionList()) <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  if (hexmesh_model != NULL)
    destroy_hexmesh_model(hexmesh_model);
  hexmesh_model = new_hexmesh_model(hexmesh_model_description);
  
  ch = 0;
  while ((hex = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(hex) == EG_HEXAHEDRON) {
      if (!EGGetGraphicsGeometry(hex, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      for (i = 1, sumd = 0; i < 8; i++) {
        VECMAC_3_VEC(v, =, pts[i], -, pts[0]);
        d = VECMAC_NORM(v);
        sumd += d;
      }
      if (user_search_tol > 0) {
	search_tol = user_search_tol;
      } else {
	d *= 0.001; search_tol = min(search_tol, d);
      }
      for (i = 0; i < 8; i++)
        nodes[i] = hexmesh_model_add_node(hexmesh_model, &pts[i], search_tol);
      hexmesh_model_add_hexvol(hexmesh_model, nodes, hex);
      ++ch;
      if (ch / 100 * 100 == ch) 
        fprintf(stderr, "Added hex #%d out of %d\n", ch, nh);
    } 
  }

 exit:
  EVUninstallHandler(v_p);
  ERptStatusMessage(1, 1, "Done");
}



static void 
set_mat_id (Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      ) {
    mat_id = TypeInGetTokenAsInt (1);
  }
}


static char *appldisp[3] = { NULL, NULL, NULL };

static int
eval_displ_at (WCRec *at, WCRec *d)
{
  int stat;
  double v;
  expeval_set_value ("x", at->x);
  expeval_set_value ("y", at->y);
  expeval_set_value ("z", at->z);
  stat = expeval_eval (appldisp[0], &v);
  if (stat != EXPEVAL_E_OK) {
    char buf[1024];
    sprintf(buf, "Failed to evaluate %s", appldisp[0]);
    ERptErrMessage(1, 2, buf, ERROR_GRADE);
    return FALSE;
  }
  d->x = v;
  stat = expeval_eval (appldisp[1], &v);
  if (stat != EXPEVAL_E_OK) {
    char buf[1024];
    sprintf(buf, "Failed to evaluate %s", appldisp[1]);
    ERptErrMessage(1, 2, buf, ERROR_GRADE);
    return FALSE;
  }
  d->y = v;
  stat = expeval_eval (appldisp[2], &v);
  if (stat != EXPEVAL_E_OK) {
    char buf[1024];
    sprintf(buf, "Failed to evaluate %s", appldisp[2]);
    ERptErrMessage(1, 2, buf, ERROR_GRADE);
    return FALSE;
  }
  d->z = v;
  return TRUE;
}

static void 
do_apply_displ (EView *v_p, caddr_t data_p)
{
  GraphicObj *g;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  while ((g = EVGetFromSelectionList()) != NULL) {
    const int maxnp = 256;
    WCRec points[maxnp];
    int i, np = maxnp;
    WCRec d;
    if (EGGetGraphicsGeometry (g, points, &np)) {
      FOR (i, np) {
        if (!eval_displ_at (&points[i], &d))
          goto exit;
        VECMAC_2_VEC (points[i], +=, d);
      }
      EGModifyGraphicsGeometry (g, points);
    }
  }
  
 exit:;
}


static void
apply_displ (Widget w, XtPointer client_data, XtPointer call_data)
{
  if (   TypeInParseLine(call_data) > 0 &&
         TypeInGetTokensTotal () >= 3
      ) {
    int i;
    FOR (i, 3) {
      if (appldisp[i]) free_instance (appldisp[i]);
      NEWP_DUPL_STRING (appldisp[i], TypeInGetTokenAsString (i+1),
                        { fprintf (stderr, "Not enough memory\n"); exit (1);
                        });
    }
    EVSetSelectActionFunction (do_apply_displ, NULL, NULL);
    EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
                     NULL, NULL,
                     NULL, NULL,
                     NULL, NULL,
                     NULL, NULL);
    ERptStatusMessage(1, 1, "Select graphics to be reshaped");
  }
}



static WCRec sweep_axis = { 0, 0, 1 };
static WCRec sweep_center = { 0, 0, 0 };

static void
set_sweep_axis (Widget w, XtPointer client_data, XtPointer call_data)
{
  if (   TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      && TypeInGetTokenType(2) == NUMBER
      && TypeInGetTokenType(3) == NUMBER
      ) {
    sweep_axis.x = TypeInGetTokenAsDouble(1);
    sweep_axis.y = TypeInGetTokenAsDouble(2);
    sweep_axis.z = TypeInGetTokenAsDouble(3);
  } else {
    ERptErrMessage(1, 1, "Bad vector components", ERROR_GRADE);
  }
}

static void
set_sweep_center (Widget w, XtPointer client_data, XtPointer call_data)
{
  if (   TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      && TypeInGetTokenType(2) == NUMBER
      && TypeInGetTokenType(3) == NUMBER
      ) {
    sweep_center.x = TypeInGetTokenAsDouble(1);
    sweep_center.y = TypeInGetTokenAsDouble(2);
    sweep_center.z = TypeInGetTokenAsDouble(3);
  } else {
    ERptErrMessage(1, 1, "Bad vector components", ERROR_GRADE);
  }
}



static void 
do_sweep_quad_to_hex (EView *v_p, caddr_t data_p)
{
  GraphicObj *q, *h;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  while ((q = EVGetFromSelectionList()) != NULL) {
    if        (   EGGraphicType(q) == EG_QUAD3D
               || EGGraphicType(q) == EG_QUADWD3D) {
      h = sweep_q_2_h(q, &sweep_center, &sweep_axis,
                      EASValsGetRotationAngle ());
      if (h != NULL) {
        EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, h);
        EMNoDrawAddGraphicsToModel(ESIModel(), h);
      }
    }
  }
  
 exit:;
}


static void
sweep_quad_to_hex_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
  EVSetSelectActionFunction(do_sweep_quad_to_hex, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select quadrilaterals to be swept");
}


static void 
do_change_mat_id (EView *v_p, caddr_t data_p)
{
  GraphicObj *hex = NULL;
  hexvol_t *h;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }
  while ((hex = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(hex) == EG_HEXAHEDRON) {
      h = hexmesh_model_hexvol_by_g_rep (hexmesh_model, hex);
      if (h == NULL) {
        ERptErrMessage(1, 2, "Hexahedron not found in model.", WARNING_GRADE);
      } else {
        hexmesh_set_hexvol_mat_id (hexmesh_model, h, mat_id);
      }
    }
  }
 exit:
  ;
}

static void 
change_mat_id_CB (Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetPickableMask (HEX_PICK_MASK);
  EVSetSelectActionFunction (do_change_mat_id, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select hexahedrons to have their mat id changed");
}



void
ESICustomize(Widget parent_pane)
{
  Arg args[3];
  int argn;
  Widget w, p;

  argn = 0;
  w = ESIAddPalette("modelling_extensions", "Modelling extensions",
                    parent_pane, args, argn, &p);
  argn = 0;
  ESIAddPopupDialog("extrusion_delta", "Extrusion vector",
                    "Vector components >",
                    "0 0 1", p, args, argn,
                    set_extrusion_vector, NULL, ESIDialogValueNumber,
                    NULL);
  argn = 0;
  w    = ESIAddButton("draw_hex_by_extrusion", "Draw hex by extrusion",
                      commandWidgetClass,
                      p, args, argn, extrude_quad_to_hex_CB, NULL);
  argn = 0;
  ESIAddPopupDialog("sweep_axis", "Sweep axis",
                    "Vector components >",
                    "0 0 1", p, args, argn,
                    set_sweep_axis, NULL, ESIDialogValueNumber,
                    NULL);
  argn = 0;
  ESIAddPopupDialog("sweep_center", "Sweep center",
                    "Vector components >",
                    "0 0 0", p, args, argn,
                    set_sweep_center, NULL, ESIDialogValueNumber,
                    NULL);
  argn = 0;
  w    = ESIAddButton("draw_hex_by_sweep", "Draw hex by sweep",
                      commandWidgetClass,
                      p, args, argn, sweep_quad_to_hex_CB, NULL);
  argn = 0;
  ESIAddPopupDialog("radius_and_nv", "Radius and vertices/side",
                    "Radius and number of vertices per side >",
                    "1 3", p, args, argn,
                    set_radius_and_nv, NULL, ESIDialogValueAny,
                    NULL);
  argn = 0;
  w    = ESIAddButton("draw_circle_quadrant", "Draw circle quadrant",
                      commandWidgetClass,
                      p, args, argn, draw_circle_quadrant_CB, NULL);
  argn = 0;
  w    = ESIAddButton("draw_sphere_octant", "Draw sphere octant",
                      commandWidgetClass,
                      p, args, argn, draw_sphere_octant_CB, NULL);
   argn = 0;
  ESIAddPopupDialog("nx_ny_nz", "Split intervals",
                    "Number of hexes per x, y, z direction >",
                    "1 1 1", p, args, argn,
                    set_split_intervals, NULL, ESIDialogValueAny,
                    NULL);
  argn = 0;
  w    = ESIAddButton("split_hex", "Split hexahedron",
                      commandWidgetClass,
                      p, args, argn, split_hex_CB, NULL);
  argn = 0;
  w    = ESIAddButton("split_quad", "Split quadrilateral",
                      commandWidgetClass,
                      p, args, argn, split_quad_CB, NULL);
  argn = 0;
  w    = ESIAddButton("load_and_eval", "Load and evaluate file",
                      commandWidgetClass,
                      p, args, argn, load_and_eval_CB, NULL);
  argn = 0;
  ESIAddPopupDialog("apply_displ", "Apply displacement",
                    "Displacement vector >",
                    "0 0 0", p, args, argn,
                    apply_displ, NULL, ESIDialogValueAny,
                    NULL);
  argn = 0;
  w    = ESIAddButton("construct_crack_surf", "Construct crack surface",
                      commandWidgetClass,
                      p, args, argn, construct_crack_surf_CB, NULL);
  argn = 0;
  w    = ESIAddButton("construct_smf", "Construct SMF",
                      commandWidgetClass,
                      p, args, argn, construct_smf_CB, NULL);
  argn = 0;
  w    = ESIAddButton("print_wasp_trac", "Print WASP tractions",
                      commandWidgetClass,
                      p, args, argn, print_wasp_trac_CB, NULL);
   argn = 0;
  w    = ESIAddButton("count_hexes", "Count hexes",
                      commandWidgetClass,
                      p, args, argn, count_hexes_CB, NULL);

 
  
  argn = 0;
  ESIAddPopupDialog("hexmesh_model_description", "Set model description",
                    "Any string >",
                    "", parent_pane, args, argn,
                    set_hexmesh_model_description, NULL, ESIDialogValueString,
                    NULL);
  
  argn = 0;
  ESIAddPopupDialog("set_user_tol", "Set search tol",
                    "# > 0",
                    "0.0000", parent_pane, args, argn,
                    set_search_tol, NULL, ESIDialogValueNumber,
                    NULL);
  
  argn = 0;
  ESIAddPopupDialog("set_rand_pos_ampl", "Set random motion amplitude",
                    "# > 0",
                    "0", parent_pane, args, argn,
                    set_rand_pos_ampl, NULL, ESIDialogValueNumber,
                    NULL);
  
  argn = 0;
  w    = ESIAddButton("make_hexmesh_model", "Make model",
                      commandWidgetClass,
                      parent_pane, args, argn, make_hexmesh_model_CB, NULL);

  argn = 0;
  w    = ESIAddButton("change_to_efg", "Change to EFG",
                      commandWidgetClass,
                      parent_pane, args, argn, change_to_efg_CB, NULL);

  argn = 0;
  ESIAddPopupDialog("diag_length_mult", "Set diagonal length multiplier",
                    "multiplier (>= 0) >",
                    "1", parent_pane, args, argn,
                    set_diag_length_mult, NULL, ESIDialogValueNumber, NULL);
  
  argn = 0;
  w    = ESIAddButton("compute_d_ms", "Compute d_m's",
                      commandWidgetClass,
                      parent_pane, args, argn, compute_d_ms_CB, NULL);

  argn = 0;
  w    = ESIAddButton("randomize_diagonals", "Randomize diagonals",
                      commandWidgetClass,
                      parent_pane, args, argn, randomize_diagonals_CB, NULL);

  argn = 0;
  w    = ESIAddButton("rand_pos", "Randomize node positions",
                      commandWidgetClass,
                      parent_pane, args, argn, rand_pos_CB, NULL);

  argn = 0;
  ESIAddPopupDialog("mat_id", "Material identifier",
                    "Mat id (> 0) >",
                    "1", parent_pane, args, argn,
                    set_mat_id, NULL, ESIDialogValueNumber, NULL);

  argn = 0;
  w    = ESIAddButton("change_mat_id", "Change material id",
                      commandWidgetClass,
                      parent_pane, args, argn, change_mat_id_CB, NULL);

  argn = 0;
  ESIAddPopupDialog("material_props", "Material properties",
                    "E nu rho (E > 0, 0 < nu <= 0.5, rho > 0) >",
                    "210000.0   0.3  7.85", parent_pane, args, argn,
                    set_E_nu, NULL, ESIDialogValueNumber, NULL);
  
  argn = 0;
  ESIAddPopupDialog("quadrature_order", "Quadrature order",
                    "Order (> 0) >",
                    "2", parent_pane, args, argn,
                    set_quadrature_order, NULL, ESIDialogValueNumber, NULL);
  
  argn = 0;
  w    = ESIAddButton("change_quadrature_order", "Change quadrature order",
                      commandWidgetClass,
                      parent_pane, args, argn, change_quadrature_CB, NULL);

  argn = 0;
  ESIAddPopupDialog("subsplit", "Sub-split into",
                    "Number of sub-cells per side (> 1) >",
                    "2", parent_pane, args, argn,
                    set_subsplit, NULL, ESIDialogValueNumber, NULL);
  
  argn = 0;
  w    = ESIAddButton("change_subsplit", "Change sub-split",
                      commandWidgetClass,
                      parent_pane, args, argn, change_subsplit_CB, NULL);

  argn = 0;
  w = ESIAddPalette("add_constraint", "Add constraint",
                    parent_pane, args, argn, &p);
  argn = 0;
  ESIAddPopupDialog("constraint_type", "Constraint type",
                    "Type [one of X, Y, Z (global dirs), A (all), F (completely free)] >",
                    "A", p, args, argn,
                    set_constraint_type, NULL, ESIDialogValueString, NULL);
  argn = 0;
  ESIAddPopupDialog("prescribed_displ_function_x", "Prescribed displacement u",
                    "Displacement u (function of x,y,z) >",
                    "0", p, args, argn,
                    set_function, (XtPointer)'x',
                    ESIDialogValueAny, NULL);
  argn = 0;
  ESIAddPopupDialog("prescribed_displ_function_x", "Prescribed displacement v",
                    "Displacement v (function of x,y,z) >",
                    "0", p, args, argn,
                    set_function, (XtPointer)'y',
                    ESIDialogValueAny, NULL);
  argn = 0;
  ESIAddPopupDialog("prescribed_displ_function_x", "Prescribed displacement w",
                    "Displacement w (function of x,y,z) >",
                    "0", p, args, argn,
                    set_function, (XtPointer)'z',
                    ESIDialogValueAny, NULL);
  w    = ESIAddButton("make_constraint", "Make constraint",
                      commandWidgetClass,
                      p, args, argn, make_constraint_CB, NULL);

  argn = 0;
  w = ESIAddPalette("add_surf_tractions", "Add surface tractions",
                    parent_pane, args, argn, &p);
  argn = 0;
  ESIAddPopupDialog("traction_function_x", "Traction T_x",
                    "Traction in the X direction (function of x,y,z) >",
                    "0", p, args, argn,
                    set_function, (XtPointer)'x',
                    ESIDialogValueAny, NULL);
  argn = 0;
  ESIAddPopupDialog("traction_function_y", "Traction T_y",
                    "Traction in the Y direction (function of x,y,z) >",
                    "0", p, args, argn,
                    set_function, (XtPointer)'y',
                    ESIDialogValueAny, NULL);
  argn = 0;
  ESIAddPopupDialog("traction_function_z", "Traction T_z",
                    "Traction in the Z direction (function of x,y,z) >",
                    "0", p, args, argn,
                    set_function, (XtPointer)'z',
                    ESIDialogValueAny, NULL);
  w    = ESIAddButton("make_conc_load", "Make surface tractions",
                      commandWidgetClass,
                      p, args, argn, make_surf_tractions_CB, NULL);


  argn = 0;
  w = ESIAddPalette("add_conc_load", "Add concentrated load",
                    parent_pane, args, argn, &p);
  argn = 0;
  ESIAddPopupDialog("conc_load_value", "Concentrated load components",
                    "Magnitude >",
                    "1.0 1.0 1.0", p, args, argn,
                    set_conc_load_components, NULL,
                    ESIDialogValueNumber, NULL);
  w    = ESIAddButton("make_conc_load", "Make concentrated load",
                      commandWidgetClass,
                      p, args, argn, make_conc_load_CB, NULL);

  ESIAddPopdownMenu("add_output_request", "Add output request",
                        parent_pane, args, argn, &w);
  ESIAddMenuItem("add_output_request_pick_node", "Pick node",
                 w, args, argn, make_output_request_CB, NULL);
  ESIAddMenuItem("add_output_request_loc", "By node location",
                 w, args, argn, make_output_request_by_loc_CB, NULL);

  argn = 0;
  w = ESIAddButton("toggle_symbolic_tracs",
                   "Print symbolic tractions", toggleWidgetClass,
                   parent_pane, args, argn, toggle_symbolic_tracs_CB,
                   (XtPointer)0);
  
  argn = 0;
  w = ESIAddButton("toggle_nodes_at_hex_centers",
                   "Generate nodes at hex centers", toggleWidgetClass,
                   parent_pane, args, argn, toggle_nodes_at_hex_centers_CB,
                   (XtPointer)0);
  
  argn = 0;
  w    = ESIAddButton("print_hexmesh_model", "Output model for WASP",
                      commandWidgetClass,
                      parent_pane, args, argn, print_hexmesh_model_CB, NULL);

  argn = 0;
  w    = ESIAddButton("print_hexmesh_tees_model", "Output model for TEES",
                      commandWidgetClass,
                      parent_pane, args, argn, print_tees_model_CB, NULL);

  argn = 0;
  w    = ESIAddButton("print_hexmesh_ndot_model", "Output model for NDOT",
                      commandWidgetClass,
                      parent_pane, args, argn, print_ndot_model_CB, NULL);

  argn = 0;
  w = ESIAddPalette("parallel_model", "Parallelization",
                    parent_pane, args, argn, &p);
  argn = 0;
  ESIAddPopupDialog("set_partition", "Set current partition",
                    "Partition number (>= 1) >",
                    "1", p, args, argn,
                    set_partition, NULL, ESIDialogValueNumber, NULL);
  
  argn = 0;
  w    = ESIAddButton("add_to_partition", "Add elem to current partition",
                      commandWidgetClass,
                      p, args, argn, add_to_partition_CB, NULL);

  argn = 0;
  w    = ESIAddButton("print_hexmesh_model", "Output model for PVM WASP",
                      commandWidgetClass,
                      p, args, argn,
                      print_hexmesh_model_parallel_CB, NULL);
}


static void
print_hexmesh_model_CB(Widget w, XtPointer ptr, XtPointer call)
{
  FILE *fp;
  char *name;
    
  fp = ESISelFile("Output file name", "OK", "CANCEL", NULL, NULL, "w", &name);
  if (fp != NULL) {
    fclose(fp);
    fp = fopen(name, "w");
    hexmesh_model_set_properties(hexmesh_model,
                               material_E, material_nu, material_rho);
    hexmesh_set_model_quadrature_order(hexmesh_model, quadrature_order);
    hexmesh_model_print(hexmesh_model, fp);
    free(name);
    fclose(fp);
  }
}


static void
print_tees_model_CB(Widget w, XtPointer ptr, XtPointer call)
{
  FILE *fp;
  char *name;
    
  fp = ESISelFile("Output file name", "OK", "CANCEL", NULL, NULL, "w", &name);
  if (fp != NULL) {
    fclose(fp);
    fp = fopen(name, "w");
    hexmesh_model_set_properties(hexmesh_model,
                               material_E, material_nu, material_rho);
    hexmesh_model_print_TEES(hexmesh_model, fp,
                             symbolic_tractions, nodes_at_hex_centers);
    free(name);
    fclose(fp);
  }
}


  
static void
print_ndot_model_CB(Widget w, XtPointer ptr, XtPointer call)
{
  FILE *fp;
  char *name;
    
  fp = ESISelFile("Output file name", "OK", "CANCEL", NULL, NULL, "w", &name);
  if (fp != NULL) {
    fclose(fp);
    fp = fopen(name, "w");
    hexmesh_model_set_properties(hexmesh_model,
                               material_E, material_nu, material_rho);
    hexmesh_model_print_NDOT(hexmesh_model, fp);
    free(name);
    fclose(fp);
  }
}


  
static void
print_hexmesh_model_parallel_CB(Widget w, XtPointer ptr, XtPointer call)
{
  FILE *fp;
  char *name;
    
  fp = ESISelFile("Output file name", "OK", "CANCEL", NULL, NULL, "w", &name);
  if (fp != NULL) {
    fclose(fp);
    fp = fopen(name, "w");
    hexmesh_model_set_properties(hexmesh_model,
                               material_E, material_nu, material_rho);
    hexmesh_model_print_parallel(hexmesh_model, name);
    free(name);
    fclose(fp);
  }
}


  
static void 
set_conc_load_components(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0 && TypeInGetTokenType(1) == NUMBER) {
    conc_load_components.x = TypeInGetTokenAsDouble(1);
  }
  if (TypeInParseLine(call_data) > 0 && TypeInGetTokenType(2) == NUMBER) {
    conc_load_components.y = TypeInGetTokenAsDouble(2);
  }
  if (TypeInParseLine(call_data) > 0 && TypeInGetTokenType(3) == NUMBER) {
    conc_load_components.z = TypeInGetTokenAsDouble(3);
  }
  
}


static void 
set_function(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  int stat;
  
  if ((stat = hexmesh_set_function((int)text_ptr, call_data)) != EXPEVAL_E_OK) {
    char buf[256];
    if (stat >= 0 && stat < EXPEVAL_E_SENTINEL)
      sprintf(buf, "Failed to evaluate %s. %s",
              (char *)call_data, expeval_err_msgs[stat]);
    else
      sprintf(buf, "Failed to evaluate %s.  Unknown error", (char *)call_data);
    ERptErrMessage(1, 2, buf, ERROR_GRADE);
  }
}

 
static void 
change_to_efg_CB(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetPickableMask(HEX_PICK_MASK);
  EVSetSelectActionFunction(do_change_to_efg, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select hexahedrons to be changed to EFG");
}


static void 
change_quadrature_CB(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetPickableMask(HEX_PICK_MASK);
  EVSetSelectActionFunction(do_change_quadrature, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select hexahedrons to be changed quadrature order");
}


static void 
do_change_subsplit(EView *v_p, caddr_t data_p);


static void 
change_subsplit_CB(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetPickableMask(HEX_PICK_MASK);
  EVSetSelectActionFunction(do_change_subsplit, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select hexahedrons to be changed sub-split");
}


static void 
compute_d_ms_CB(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  ERptStatusMessage(1, 1, "Computing d_m for all nodes");
  hexmesh_compute_d_m(hexmesh_model, diag_length_mult);
  ERptStatusMessage(1, 1, "Done");
}



static void 
do_change_to_efg(EView *v_p, caddr_t data_p)
{
  GraphicObj *hex = NULL;
  hexvol_t *h;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }
  while ((hex = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(hex) == EG_HEXAHEDRON) {
      h = hexmesh_model_hexvol_by_g_rep(hexmesh_model, hex);
      if (h == NULL) {
        ERptErrMessage(1, 2, "Hexahedron not found in model.", WARNING_GRADE);
      } else {
        hexmesh_model_change_to_efg(hexmesh_model, h);
      }
    }
  }
 exit:
  ERptStatusMessage(1, 1, "Done");
}



static void 
do_change_quadrature(EView *v_p, caddr_t data_p)
{
  GraphicObj *hex = NULL;
  hexvol_t *h;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }
  while ((hex = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(hex) == EG_HEXAHEDRON) {
      h = hexmesh_model_hexvol_by_g_rep(hexmesh_model, hex);
      if (h == NULL) {
        ERptErrMessage(1, 2, "Hexahedron not found in model.", WARNING_GRADE);
      } else {
        hexmesh_set_hexvol_quadrature_order(hexmesh_model, h,
                                            quadrature_order);
      }
    }
  }
 exit:
  ;
}




static void 
do_change_subsplit(EView *v_p, caddr_t data_p)
{
  GraphicObj *hex = NULL;
  hexvol_t *h;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }
  while ((hex = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(hex) == EG_HEXAHEDRON) {
      h = hexmesh_model_hexvol_by_g_rep(hexmesh_model, hex);
      if (h == NULL) {
        ERptErrMessage(1, 2, "Hexahedron not found in model.", WARNING_GRADE);
      } else {
        hexmesh_set_hexvol_subsplit(hexmesh_model, h, subsplit);
      }
    }
  }
 exit:
  ;
}


static void 
do_add_to_partition(EView *v_p, caddr_t data_p);


static void 
add_to_partition_CB(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetPickableMask(HEX_PICK_MASK);
  EVSetSelectActionFunction(do_add_to_partition, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select hexahedrons to add to current partition");
}

static void 
do_add_to_partition(EView *v_p, caddr_t data_p)
{
  GraphicObj *hex = NULL;
  hexvol_t *h;
  int hp;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }
  while ((hex = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(hex) == EG_HEXAHEDRON) {
      h = hexmesh_model_hexvol_by_g_rep(hexmesh_model, hex);
      if (h == NULL) {
        ERptErrMessage(1, 2, "Hexahedron not found in model.", WARNING_GRADE);
      } else {
        hp = hexmesh_get_hexvol_partition(hexmesh_model, h);
        hexmesh_set_hexvol_partition(hexmesh_model, h, curr_partition);
      }
    }
  }
 exit:
  ;
}

static void 
set_E_nu(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      && TypeInGetTokenType(2) == NUMBER
      && TypeInGetTokenType(3) == NUMBER
      ) {
    material_E   = TypeInGetTokenAsDouble(1);
    material_nu  = TypeInGetTokenAsDouble(2);
    material_rho = TypeInGetTokenAsDouble(3);
    if (material_E <= 0) 
      ERptErrMessage(1, 1, "Bad elasticity modulus", ERROR_GRADE);
    if (material_nu <= 0 || material_nu > 0.5) 
      ERptErrMessage(1, 1, "Bad Poisson constant", ERROR_GRADE);
    if (material_rho <= 0) 
      ERptErrMessage(1, 1, "Bad mass density", ERROR_GRADE);
  }
}



static void 
set_diag_length_mult(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      ) {
    diag_length_mult = TypeInGetTokenAsDouble(1);
    if (diag_length_mult <= 0) {
      ERptErrMessage(1, 1, "Bad diag_length_mult;  Must be > 0", ERROR_GRADE);
      diag_length_mult = 1;
    }
  }
}


static void 
set_search_tol(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      ) {
    user_search_tol = TypeInGetTokenAsDouble(1);
    if (user_search_tol <= 0) {
      ERptErrMessage(1, 1, "Bad search tolerance;  Must be > 0",
		     ERROR_GRADE);
      user_search_tol = 0;
    }
  }
}

static void 
set_rand_pos_ampl (Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      ) {
    rand_pos_ampl = TypeInGetTokenAsDouble(1);
    if (rand_pos_ampl <= 0) {
      ERptErrMessage(1, 1, "Bad amplitude;  Must be > 0",
		     ERROR_GRADE);
      rand_pos_ampl = 0;
    }
  }
}


static void 
set_quadrature_order(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      ) {
    quadrature_order   = TypeInGetTokenAsInt(1);
    if (quadrature_order <= 0) 
      ERptErrMessage(1, 1, "Bad quadrature order", ERROR_GRADE);
    if (hexmesh_model)
      hexmesh_set_model_quadrature_order(hexmesh_model, quadrature_order);
  }
}


static void 
set_subsplit(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      ) {
    subsplit   = TypeInGetTokenAsInt(1);
    if (subsplit <= 1) 
      ERptErrMessage(1, 1, "Bad subsplit (should be > 1)", ERROR_GRADE);
  }
}


static void 
set_partition(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      ) {
    curr_partition   = TypeInGetTokenAsInt(1);
    if (curr_partition < 1) 
      ERptErrMessage(1, 1, "Bad partition number", ERROR_GRADE);
  }
}


 
static void 
set_hexmesh_model_description(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0 && TypeInGetTokenType(1) == STRING) {
    if (hexmesh_model_description)
      free(hexmesh_model_description);
    DUPLICATE_STRING(hexmesh_model_description, TypeInGetTokenAsString(1));
  }
}


 
static void 
set_constraint_type(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0 && TypeInGetTokenType(1) == STRING) {
    constraint_type = *(TypeInGetTokenAsString(1));
    if (   constraint_type != 'x'
        && constraint_type != 'X'
        && constraint_type != 'y'
        && constraint_type != 'Y'
        && constraint_type != 'z'
        && constraint_type != 'Z'
        && constraint_type != 'a'
        && constraint_type != 'A'
        && constraint_type != 'f'
        && constraint_type != 'F'
        )
      ERptErrMessage(1, 1, "Bad constraint type", ERROR_GRADE);
  }
}



static void
make_constraint_CB(Widget w, XtPointer ptr, XtPointer call)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetPickableMask(NODE_PICK_MASK);
  EVSetSelectActionFunction(do_make_constraint, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1,
                    "Select nodes to be included in the constraint");
}


static void
rand_pos_CB (Widget w, XtPointer ptr, XtPointer call)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetPickableMask(NODE_PICK_MASK);
  EVSetSelectActionFunction(do_rand_pos, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1,
                    "Select nodes to be included in the constraint");
}


static WCRec extrusion_vector = { 0, 0, 1 };

static void
set_extrusion_vector(Widget w, XtPointer client_data, XtPointer call_data)
{
  if (   TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      && TypeInGetTokenType(2) == NUMBER
      && TypeInGetTokenType(3) == NUMBER
      ) {
    extrusion_vector.x = TypeInGetTokenAsDouble(1);
    extrusion_vector.y = TypeInGetTokenAsDouble(2);
    extrusion_vector.z = TypeInGetTokenAsDouble(3);
  } else {
    ERptErrMessage(1, 1, "Bad vector components", ERROR_GRADE);
  }
}




static void 
do_extrude_quad_to_hex(EView *v_p, caddr_t data_p);

static void
extrude_quad_to_hex_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
  EVSetSelectActionFunction(do_extrude_quad_to_hex, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select quadrilaterals to be extruded");
}


static void 
do_split_hex(EView *v_p, caddr_t data_p);


static void
split_hex_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
  ESIHandleCmd ("LAYER OFF 1313");
  EVSetSelectActionFunction(do_split_hex, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select hexahedra to split");
}



static void 
do_split_quad(EView *v_p, caddr_t data_p);


static void
split_quad_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
  ESIHandleCmd ("LAYER OFF 1313");
  EVSetSelectActionFunction(do_split_quad, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select quadrilaterals to split");
}




static void 
do_print_wasp_trac(EView *v_p, caddr_t data_p);


static void
print_wasp_trac_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
  EVSetSelectActionFunction(do_print_wasp_trac, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select quadrilaterals to use");
}





static void
make_conc_load_CB(Widget w, XtPointer ptr, XtPointer call)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetPickableMask(NODE_PICK_MASK);
  EVSetSelectActionFunction(do_make_conc_load, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1,
                    "Select nodes to be assigned the concentrated load");
}


static void
make_surf_tractions_CB(Widget w, XtPointer ptr, XtPointer call)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetPickableMask(NODE_PICK_MASK);
  EVSetSelectActionFunction(do_make_surf_tractions, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1,
                    "Select nodes to be assigned the tractions");
}


static void
make_output_request_CB(Widget w, XtPointer ptr, XtPointer call)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetPickableMask(NODE_PICK_MASK);
  EVSetSelectActionFunction(do_make_output_request, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1,
                    "Select nodes to be included in the output");
}




static void
make_output_request_by_loc_CB(Widget w, XtPointer ptr, XtPointer call)
{
  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  EVSetApplyToViewFunction(do_make_output_request_by_loc,
                           NULL, NULL);
  EMPushHandler(ESIModel(), EVApplyToViewHandler, NULL);
  
  ERptStatusMessage(1, 1,
                    "Pick points included in the output");
}


static void 
do_make_output_request_by_loc(EView *v_p, caddr_t data_p, WCRec *p)
{
  node_t *n;

  n = hexmesh_model_find_node(hexmesh_model, p, search_tol);
  if (n == NULL) {
    ERptErrMessage(1, 2, "No node at this location", WARNING_GRADE);
  } else {
    hexmesh_model_add_output_request(hexmesh_model, n);
  }
}



static void 
do_make_constraint(EView *v_p, caddr_t data_p)
{
  GraphicObj *node = NULL;
  WCRec pts[8];
  int np;
  node_t *n;
  LIST nodes = NULL;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  nodes = make_list();
  
  while ((node = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(node) == EG_MARKER3D) {
      if (!EGGetGraphicsGeometry(node, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      n = hexmesh_model_find_node(hexmesh_model, &pts[0], search_tol);
      if (n == NULL) {
        ERptStatusMessage(1, 1, "Bad graphic object -- no associated node");
        goto exit;
      }
      add_to_tail(nodes, n);
    } 
  }

  hexmesh_model_add_constraint(hexmesh_model, constraint_type, nodes);

 exit:
  if (nodes)
    free_list(nodes, NOT_NODES);
  ERptStatusMessage(1, 1, "Done");
}


static void 
do_rand_pos (EView *v_p, caddr_t data_p)
{
  GraphicObj *node = NULL;
  WCRec pts[8];
  WCRec l;
  int np;
  node_t *n;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  while ((node = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(node) == EG_MARKER3D) {
      if (!EGGetGraphicsGeometry(node, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      n = hexmesh_model_find_node (hexmesh_model, &pts[0], search_tol);
      if (n == NULL) {
        ERptStatusMessage(1, 1, "Bad graphic object -- no associated node");
        goto exit;
      }
      n->loc.x += rand_pos_ampl * rand_0_to_1 ();
      n->loc.y += rand_pos_ampl * rand_0_to_1 ();
      n->loc.z += rand_pos_ampl * rand_0_to_1 ();
      EGModifyGraphicsGeometry (node, &n->loc);
    } 
  }

 exit:
  ERptStatusMessage(1, 1, "Done");
}

static void 
do_make_surf_tractions(EView *v_p, caddr_t data_p)
{
  GraphicObj *node = NULL;
  WCRec pts[8];
  int np;
  node_t *n;
  LIST nodes = NULL;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  nodes = make_list();
  
  while ((node = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(node) == EG_MARKER3D) {
      if (!EGGetGraphicsGeometry(node, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      n = hexmesh_model_find_node(hexmesh_model, &pts[0], search_tol);
      if (n == NULL) {
        ERptStatusMessage(1, 1, "Bad graphic object -- no associated node");
        goto exit;
      }
      add_to_tail(nodes, n);
    } 
  }

  hexmesh_model_add_surf_tractions(hexmesh_model, nodes);

 exit:
  if (nodes)
    free_list(nodes, NOT_NODES);
  ERptStatusMessage(1, 1, "Done");
}



static void 
do_make_conc_load(EView *v_p, caddr_t data_p)
{
  GraphicObj *node = NULL;
  WCRec pts[8];
  int np;
  node_t *n;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  while ((node = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(node) == EG_MARKER3D) {
      if (!EGGetGraphicsGeometry(node, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      n = hexmesh_model_find_node(hexmesh_model, &pts[0], search_tol);
      if (n == NULL) {
        ERptStatusMessage(1, 1, "Bad graphic object -- no associated node");
      } else
        hexmesh_model_add_conc_load(hexmesh_model, n, &conc_load_components);
    } 
  }

 exit:
  ERptStatusMessage(1, 1, "Done");
}



static void 
do_make_output_request(EView *v_p, caddr_t data_p)
{
  GraphicObj *node = NULL;
  WCRec pts[8];
  int np;
  node_t *n;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  while ((node = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(node) == EG_MARKER3D) {
      if (!EGGetGraphicsGeometry(node, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      n = hexmesh_model_find_node(hexmesh_model, &pts[0], search_tol);
      if (n == NULL) {
        ERptStatusMessage(1, 1, "Bad graphic object -- no associated node");
      } else
        hexmesh_model_add_output_request(hexmesh_model, n);
    } 
  }

 exit:
  ERptStatusMessage(1, 1, "Done");
}


static void 
do_extrude_quad_to_hex(EView *v_p, caddr_t data_p)
{
  GraphicObj *q, *h;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  while ((q = EVGetFromSelectionList()) != NULL) {
    if        (   EGGraphicType(q) == EG_QUAD3D
               || EGGraphicType(q) == EG_QUADWD3D) {
      h = extrude_q_2_h(q, &extrusion_vector);
      if (h != NULL) {
        EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, h);
        EMNoDrawAddGraphicsToModel(ESIModel(), h);
      }
    }
  }
  
 exit:;
}


static void 
do_split_hex(EView *v_p, caddr_t data_p)
{
  GraphicObj *h;
  GraphicObj *new_h;
  LIST_OF(GraphicObj *) newhexes;
  int currlay;
  int totnh = 0;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  totnh = 0;
  while ((h = EVGetFromSelectionList()) != NULL) {
    if        (   EGGraphicType(h) == EG_HEXAHEDRON
               || EGGraphicType(h) == EG_HEXAHEDRONWD) {
      newhexes = split_h(h, nx, ny, nz);
      currlay = EASValsGetLayer();
      EASValsSetLayer (1313);
      EGWithMaskChangeAttributes(LAYER_MASK, h);
      EASValsSetLayer(currlay);
      new_h = NULL;
      while ((new_h = (GraphicObj *)get_list_next(newhexes, new_h)) != NULL) {
	EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, new_h);
	EMNoDrawAddGraphicsToModel(ESIModel(), new_h);
	totnh++;
      }
    }
  }

  
 exit:
  {
    char buf[64];
    sprintf(buf, "Total of new hexes: %d", totnh);
    ERptStatusMessage(1, 1, buf);
  }
}




static void 
do_split_quad(EView *v_p, caddr_t data_p)
{
  GraphicObj *q;
  GraphicObj *new_q;
  LIST_OF(GraphicObj *) newquads;
  int currlay;
  int totnq = 0;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  totnq = 0;
  while ((q = EVGetFromSelectionList()) != NULL) {
    if        (   EGGraphicType(q) == EG_QUAD3D
               || EGGraphicType(q) == EG_QUADWD3D) {
      newquads = split_q(q, nx, ny, nz);
      currlay = EASValsGetLayer();
      EASValsSetLayer (1313);
      EGWithMaskChangeAttributes(LAYER_MASK, q);
      EASValsSetLayer(currlay);
      new_q = NULL;
      while ((new_q = (GraphicObj *)get_list_next(newquads, new_q)) != NULL) {
	EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, new_q);
	EMNoDrawAddGraphicsToModel(ESIModel(), new_q);
	totnq++;
      }
    }
  }

  
 exit:
  {
    char buf[64];
    sprintf(buf, "Total of new quads: %d", totnq);
    ERptStatusMessage(1, 1, buf);
  }
}


static void
toggle_symbolic_tracs_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  Boolean s;
  Arg al[1];
  int ac;
  
  ac = 0;
  XtSetArg(al[ac], XtNstate, &s); ac++;
  XtGetValues(w, al, ac);
  symbolic_tractions = s;
}


static void
toggle_nodes_at_hex_centers_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  Boolean s;
  Arg al[1];
  int ac;
  
  ac = 0;
  XtSetArg(al[ac], XtNstate, &s); ac++;
  XtGetValues(w, al, ac);
  nodes_at_hex_centers = s;
}

static void
load_and_eval_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
#define MAX_LINE_LEN 256
  char linebuf[MAX_LINE_LEN];
  FILE *stream = NULL;
  char *name;
  int stat;
  double x;
#define STRIP_NEWLINE()                                         \
  {                                                             \
    int _i = 0;                                                 \
    while (_i < MAX_LINE_LEN) {                                 \
      if (linebuf[_i] == '\n') { linebuf[_i] = ' '; break; }    \
      _i++;                                                     \
    }                                                           \
  }

  stream = ESISelFile("Select file name",
                  "   OK   ", "CANCEL", NULL, NULL, "r", &name);
  if (stream) {
    fclose(stream);
  } else {
    return; /* Cancelled */
  }
  
  stream = fopen(name, "r");
  if (stream) {
    errno = 0;
    while (fgets(linebuf, MAX_LINE_LEN-1, stream) != NULL) {
      STRIP_NEWLINE();
      stat = expeval_eval(linebuf, &x);
      if (stat != EXPEVAL_E_OK) {
        char buf[1024];
        sprintf(buf, "Failed to evaluate %s", linebuf);
        ERptErrMessage(1, 2, buf, ERROR_GRADE);
      }
    }
    if (ferror(stream)) {
      fprintf(stderr, "File %s: %s", name, strerror(errno));
    }
    if (!feof(stream)) {
      fprintf(stderr, "Bad command file\n");
    }
    fclose(stream);
  }
}



static void
randomize_diagonals_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  EPixel color;

  if (hexmesh_model == NULL) {
    ERptErrMessage(1, 2, "Must create model first.", ERROR_GRADE);
    return;
  }
  
  color = EASValsGetColor();
  ESIHandleCmd("COLOR CYAN");

  hexmesh_randomize_06_diagonal(hexmesh_model);
  EASValsSetColor(color);
}


static void 
do_construct_crack_surf(EView *v_p, caddr_t data_p);

static void
construct_crack_surf_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  crack_surf_format = CRACK_SURF_FORMAT;
  EVSetSelectActionFunction(do_construct_crack_surf, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);

  ERptStatusMessage(1, 1, "Select triangles to be included in the surface");
}

static void
construct_smf_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  crack_surf_format = SMF_FORMAT;
  EVSetSelectActionFunction(do_construct_crack_surf, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);

  ERptStatusMessage(1, 1, "Select triangles to be included in the surface");
}



static void 
do_construct_crack_surf(EView *v_p, caddr_t data_p)
{
  GraphicObj *go;
  int np;
  WCRec pts[4];
  WCRec tpnts[3];
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  crack_surf_start();
  while ((go = EVGetFromSelectionList()) != NULL) {
    if        (   EGGraphicType(go) == EG_QUAD3D
               || EGGraphicType(go) == EG_QUADWD3D) {
      if (!EGGetGraphicsGeometry(go, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      tpnts[0] = pts[0];
      tpnts[1] = pts[1];
      tpnts[2] = pts[2];
      crack_surf_add_triangle(tpnts);
      tpnts[0] = pts[2];
      tpnts[1] = pts[3];
      tpnts[2] = pts[0];
      crack_surf_add_triangle(tpnts);
    } else if (   EGGraphicType(go) == EG_TRIANGLE3D
               || EGGraphicType(go) == EG_TRIANGLEWD3D) {
      if (!EGGetGraphicsGeometry(go, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      crack_surf_add_triangle(pts);
    } else if (EGGraphicType(go) == EG_LINE3D) {
      if (!EGGetGraphicsGeometry(go, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      crack_surf_add_front_segment(pts);
    }
  }

  {
    char buf[L_tmpnam] = "./tmp";
    FILE *fp;
  
    tmpnam(buf);
    fp = fopen(buf, "w");

    crack_surf_print(fp, crack_surf_format);

    fclose(fp);
    ESIPopupInfoFromFile(buf);  
  }
  
 exit:;
}



static double radius = 1;
static int nv_per_side = 2;

static void
set_radius_and_nv(Widget w, XtPointer client_data, XtPointer call_data)
{
  if (   TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      && TypeInGetTokenType(2) == NUMBER
      ) {
    radius      = TypeInGetTokenAsDouble(1);
    nv_per_side = TypeInGetTokenAsInt(2);
  } else {
    ERptErrMessage(1, 1, "Bad radius or number of vertices per side",
		   ERROR_GRADE);
  }
  if (radius <= 0)
      ERptErrMessage(1, 1, "Bad radius (should be >0)", ERROR_GRADE);
  if (nv_per_side < 2)
      ERptErrMessage(1, 1,
		     "Bad  or number of vertices per side (should be >= 2)",
		     ERROR_GRADE);
}


static void
draw_circle_quadrant_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
    if (!generate_circle_quadrant(nv_per_side, radius))
	ERptErrMessage(1, 1, "Generation failed", ERROR_GRADE);
}

static void
draw_sphere_octant_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
    if (!generate_sphere_octant(nv_per_side, radius))
	ERptErrMessage(1, 1, "Generation failed", ERROR_GRADE);
}




static void
set_split_intervals(Widget w, XtPointer client_data, XtPointer call_data)
{
  if (   TypeInParseLine(call_data) > 0
      ) {
    if (TypeInGetTokenType(1) == NUMBER)
      nx =  TypeInGetTokenAsInt(1);
    if (TypeInGetTokenType(2) == NUMBER)
      ny =  TypeInGetTokenAsInt(2);
    if (TypeInGetTokenType(3) == NUMBER)
      nz =  TypeInGetTokenAsInt(3);
  } else {
    ERptErrMessage(1, 1, "Bad number of subdivisions",
		   ERROR_GRADE);
  }
}


static void 
do_count_hexes(EView *v_p, caddr_t data_p);

static void
count_hexes_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  EVSetSelectActionFunction(do_count_hexes, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select objects to be counted");
}


static void 
do_count_hexes(EView *v_p, caddr_t data_p)
{
  GraphicObj *g = NULL;
  int count = 0;
  char buf[132];
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }
  count = 0;
  while ((g = EVGetFromSelectionList()) != NULL) {
    if (EGGraphicType(g) == EG_HEXAHEDRON)
      count++;
  }
 exit:
  sprintf(buf, "%d hexahedra selected", count);
  ERptStatusMessage(1, 1, buf);
}



static void 
do_print_wasp_trac(EView *v_p, caddr_t data_p)
{
  GraphicObj *q;
  char buf[L_tmpnam] = "./tmp";
  FILE *fp;
  int np;
  WCRec pts[4];
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  tmpnam(buf);
  fp = fopen(buf, "w");

  fprintf(fp, "DISTRIBL\n");
  
  while ((q = EVGetFromSelectionList()) != NULL) {
    if        (   EGGraphicType(q) == EG_QUAD3D
               || EGGraphicType(q) == EG_QUADWD3D) {
      if (!EGGetGraphicsGeometry(q, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      fprintf(fp, "1 DEAD_PRESSURE 1\n");
      fprintf(fp, "%g %g %g ???\n", pts[0].x, pts[0].y, pts[0].z); 
      fprintf(fp, "%g %g %g ???\n", pts[1].x, pts[1].y, pts[1].z); 
      fprintf(fp, "%g %g %g ???\n", pts[2].x, pts[2].y, pts[2].z); 
      fprintf(fp, "%g %g %g ???\n", pts[3].x, pts[3].y, pts[3].z);
    }
  }

  fprintf(fp, "END DISTRIBL\n");

  fclose(fp);
  ESIPopupInfoFromFile(buf);  

 exit:
  ;
}

