#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#include "tfmap.h"


typedef struct vect {
  double x, y;
}   vect;



/* 
c89 -D_HPUX_SOURCE -g  bcurv.c tfmap.c circgen.c -lm
 */

/* Modify this */
#ifndef NV_PER_SIDE
#   define NV_PER_SIDE  17
#endif
#ifndef RADIUS
#   define RADIUS      1.0
#endif



#define STEP 1.0 / (double)(NV_PER_SIDE - 1)

/*
     p5                    


                          p6


     p3            p4           





      p0            p1           p2
*/
static vect p[7] = {
  {0.0, 0.0},
  {0.5, 0.0},
  {1.0, 0.0},
  {0.0, 0.5},
  {0.42, 0.42},
  {0.0, 1.0},
  {0.70710678118654752440, 0.70710678118654752440}
};


#define One 1.001


static void
q4_N(double ksi, double eta, double n[4]);

#define v3_scale(v, s) { (v)->x *= s; (v)->y *= s; }


int
main(void)
{
  double ksi, eta, N[4];
  vect v;
  int nv, np, nr, nc, i, j, n1, n2, n3, n4;
  tfmap_t *tfm;
  b_curve_t c0, c1, c2, c3;
#define PT(num) p[num].x, p[num].y
#define CHK(p)                                                           \
  if ((p) == NULL) {                                                     \
    fprintf(stderr, "Failed " #p " in %s (l. %d)\n", __FILE__, __LINE__);\
    exit(1);                                                             \
  }
#define DUMP()                                                  \
  {                                                             \
    printf("%d %d\n", nv, np);                                  \
                                                                \
    for (ksi = 0; ksi < One; ksi += STEP) {                     \
      for (eta = 0; eta < One; eta += STEP) {                   \
        tfmap_interpolate(tfm, ksi, eta, &v.x, &v.y);           \
        v3_scale(&v, RADIUS);                                   \
        printf("%g %g %g\n", v.x, 0.0, v.y);                    \
      }                                                         \
    }                                                           \
                                                                \
    for (i = 0; i < nr; i++)                                    \
      for (j = 0; j < nc; j++) {                                \
        n1 = i * (nc+1) + j + 1;                                \
        n2 = n1 + 1;                                            \
        n4 = n1 + (nc+1);                                       \
        n3 = n4 + 1;                                            \
        printf("4\t\t%d \t%d \t%d \t%d\n", n1, n2, n3, n4);     \
      }                                                         \
  }

  nr = nc = NV_PER_SIDE - 1;
  nv = (nr + 1) * (nc + 1);
  np = nr * nc;

  CHK(c0 = bcurv_make_line(PT(0), PT(1))); 
  CHK(c1 = bcurv_make_line(PT(3), PT(4))); 
  CHK(c2 = bcurv_make_line(PT(0), PT(3))); 
  CHK(c3 = bcurv_make_line(PT(1), PT(4))); 
  CHK(tfm = tfmap_make_tfmap(c0, c1, c2, c3, 0.001));
  DUMP();

  CHK(c0 = bcurv_make_line(PT(1), PT(2))); 
  CHK(c1 = bcurv_make_line(PT(4), PT(6))); 
  CHK(c2 = bcurv_make_line(PT(1), PT(4))); 
  CHK(c3 = bcurv_make_circ_arc(PT(2), PT(6), 0, 0, 0, PT(0), 0)); 
  CHK(tfm = tfmap_make_tfmap(c0, c1, c2, c3, 0.001));
  DUMP();

  CHK(c0 = bcurv_make_line(PT(3), PT(4))); 
  CHK(c1 = bcurv_make_circ_arc(PT(5), PT(6), 0, 0, 0, PT(0), 0)); 
  CHK(c2 = bcurv_make_line(PT(3), PT(5))); 
  CHK(c3 = bcurv_make_line(PT(4), PT(6))); 
  CHK(tfm = tfmap_make_tfmap(c0, c1, c2, c3, 0.001));
  DUMP();
  
  return 1;
}



static void
q4_N(double ksi, double eta, double n[4])
{
  double one_minus_ksi = (1 - ksi);
  double one_plus_ksi  = (1 + ksi);
  double one_minus_eta = (1 - eta);
  double one_plus_eta  = (1 + eta);
  
  n[0] = 0.25 * one_minus_ksi * one_minus_eta;
  n[1] = 0.25 * one_plus_ksi  * one_minus_eta;
  n[2] = 0.25 * one_plus_ksi  * one_plus_eta;
  n[3] = 0.25 * one_minus_ksi * one_plus_eta;
}
