#include "hexmesh.h"
#include "tfmap.h"



typedef struct vect {
  double x, y;
}   vect;



/*
     p5                    


                          p6


     p3            p4           





      p0            p1           p2
*/
static vect p[7] = {
  {0.0, 0.0},
  {0.5, 0.0},
  {1.0, 0.0},
  {0.0, 0.5},
  {0.42, 0.42},
  {0.0, 1.0},
  {0.70710678118654752440, 0.70710678118654752440}
};




static void
q4_N(double ksi, double eta, double n[4]);

#define scale(v, s) { (v)->x *= s; (v)->y *= s; }


int
generate_circle_quadrant(int nv_per_side, double radius)
{
  double ksi, eta, N[4];
  vect v[4];
  WCRec pts[4];
  int nv, np, nr, nc, i, j, n1, n2, n3, n4;
  tfmap_t *tfm;
  b_curve_t c0, c1, c2, c3;
  double step = 1.0 / (double)(nv_per_side - 1);
  double top = 1.0001 - step;
  GraphicObj *q;
#define PT(num) p[num].x, p[num].y
#define CHK(p)                                                           \
  if ((p) == NULL) {                                                     \
    fprintf(stderr, "Failed " #p " in %s (l. %d)\n", __FILE__, __LINE__);\
    return 0;                                                            \
  }
#define DUMP()								\
  {									\
    for (ksi = 0; ksi < top; ksi += step) {				\
      for (eta = 0; eta < top; eta += step) {				\
        tfmap_interpolate(tfm, ksi     ,      eta, &v[0].x, &v[0].y);	\
        tfmap_interpolate(tfm, ksi+step,      eta, &v[1].x, &v[1].y);	\
        tfmap_interpolate(tfm, ksi+step, eta+step, &v[2].x, &v[2].y);	\
        tfmap_interpolate(tfm, ksi     , eta+step, &v[3].x, &v[3].y);	\
        scale(&v[0], radius);						\
        scale(&v[1], radius);						\
        scale(&v[2], radius);						\
        scale(&v[3], radius);						\
	for (i = 0; i < 4; i++) {					\
	    pts[i].x = v[i].x;						\
	    pts[i].y = v[i].y;						\
	    pts[i].z = 0;						\
	}								\
	q = CreateQuad3D(pts);						\
	EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, q);                 \
	EMNoDrawAddGraphicsToModel(ESIModel(), q);			\
      }									\
    }									\
  }
  

  nr = nc = nv_per_side - 1;
  nv = (nr + 1) * (nc + 1);
  np = nr * nc;

  CHK(c0 = bcurv_make_line(PT(0), PT(1))); 
  CHK(c1 = bcurv_make_line(PT(3), PT(4))); 
  CHK(c2 = bcurv_make_line(PT(0), PT(3))); 
  CHK(c3 = bcurv_make_line(PT(1), PT(4))); 
  CHK(tfm = tfmap_make_tfmap(c0, c1, c2, c3, 0.001));
  DUMP();

  CHK(c0 = bcurv_make_line(PT(1), PT(2))); 
  CHK(c1 = bcurv_make_line(PT(4), PT(6))); 
  CHK(c2 = bcurv_make_line(PT(1), PT(4))); 
  CHK(c3 = bcurv_make_circ_arc(PT(2), PT(6), 0, 0, 0, PT(0), 0)); 
  CHK(tfm = tfmap_make_tfmap(c0, c1, c2, c3, 0.001));
  DUMP();

  CHK(c0 = bcurv_make_line(PT(3), PT(4))); 
  CHK(c1 = bcurv_make_circ_arc(PT(5), PT(6), 0, 0, 0, PT(0), 0)); 
  CHK(c2 = bcurv_make_line(PT(3), PT(5))); 
  CHK(c3 = bcurv_make_line(PT(4), PT(6))); 
  CHK(tfm = tfmap_make_tfmap(c0, c1, c2, c3, 0.001));
  DUMP();
  
  return 1;
}



static void
q4_N(double ksi, double eta, double n[4])
{
  double one_minus_ksi = (1 - ksi);
  double one_plus_ksi  = (1 + ksi);
  double one_minus_eta = (1 - eta);
  double one_plus_eta  = (1 + eta);
  
  n[0] = 0.25 * one_minus_ksi * one_minus_eta;
  n[1] = 0.25 * one_plus_ksi  * one_minus_eta;
  n[2] = 0.25 * one_plus_ksi  * one_plus_eta;
  n[3] = 0.25 * one_minus_ksi * one_plus_eta;
}
