
#include "hexmesh.h"

#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif


static void
update_dmaxes(hexvol_t *h, double diag_length_mult);


void
compute_d_ms(hexmesh_model_t *model, double diag_length_mult)
{
  node_t *n;
  hexvol_t *h;

  loop_through_list(model->nodes, n, node_t *) {n->d_m = 0;}

  loop_through_list(model->hexvols, h, hexvol_t *) {
    update_dmaxes(h, diag_length_mult);
  }
}


static double 
hexvol_max_diag_length(hexvol_t *h);


static void
update_dmaxes(hexvol_t *h, double diag_length_mult)
{
  int i;
  double diag_length;

  diag_length = hexvol_max_diag_length(h);
  h->d_m = diag_length * diag_length_mult;
  FOR(i, 8) {
    h->nodes[i]->d_m = max(h->nodes[i]->d_m, h->d_m);
  }
}


static double 
maxdi3(double d1, double d2, double d3);

static double 
hexvol_max_diag_length(hexvol_t *h)
{
  double d1, d2, d3;

  d1 = v3_distance(&h->nodes[0]->loc, &h->nodes[6]->loc);
  d2 = v3_distance(&h->nodes[1]->loc, &h->nodes[7]->loc);
  d3 = v3_distance(&h->nodes[2]->loc, &h->nodes[4]->loc);
  return maxdi3(d1, d2, d3);
}


static double 
maxdi2(double d1, double d2);


static double 
maxdi3(double d1, double d2, double d3)
{
  return maxdi2(d1, maxdi2(d2, d3));
}


static double 
maxdi2(double d1, double d2)
{
  return max(d1, d2);
}

