#ifndef HEXMESH_H
#   define HEXMESH_H


#define WANT_VECTP_ALIASES
#include "Esimple.h"
#include <math.h>
#include <stdlib.h>
#include "expevalP.h"
#include "vectP.h"
#define VECMAC_VECTOR_TYPE 1 
#include "vecmacP.h"
#include "listP.h"
#include "dictP.h"
#include "randP.h"
#include "newP.h"
#include "g3d.h"
#define USE_PIRAT 1
#if USE_PIRAT
#include "pirat.h"
#else
#include "prap3.h"
#endif

#define FOR(_Counter_, _Up_)                                      \
   for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++)


#define ListOf(type) LIST 

typedef struct hexmesh_model_t {
  char                           *description;
  double                          thickness;
  double                          E;
  double                          nu;
  double                          rho;
  int                             quadrature_order;
  int                             nnodes;
  ListOf(node_t *)                nodes;
  int                             nhexvols;
  ListOf(hexvol_t *)              hexvols;
  dict_dictionary                 grep_to_hexvol_dict;
  int                             nconc_loads;
  ListOf(conc_load_t *)           conc_loads;
  int                             noutput_requests;
  ListOf(output_request_t *)      output_requests;
#if USE_PIRAT
  pirat_front_end_t              *node_ss;
#else
  char                           *node_ss;
#endif
}              hexmesh_model_t;



#define CONSTRAINT_FREE  1
#define CONSTRAINT_FIXED 0


typedef enum { FE_NODE, EFG_NODE, FE_EFG_NODE } node_type_t;

#define NINCID2FE(n)  (n->incident_to_fe_hexvol)
#define NINCID2EFG(n) (n->incident_to_efg_hexvol)
#define N_INC_PURE_FE(n)  (NINCID2FE(n) && (!NINCID2EFG(n)))
#define N_INC_MIXED(n)    (NINCID2FE(n) && NINCID2EFG(n))
#define N_INC_PURE_EFG(n) ((!NINCID2FE(n)) && NINCID2EFG(n))

#define NODE_IS_PURE_FE(n)     (n->type == FE_NODE)
#define NODE_IS_PURE_EFG(n)    (n->type == EFG_NODE)
#define NODE_IS_FE_EFG(n)      (n->type == FE_EFG_NODE)
#define NODE_IS_FE(n)          (NODE_IS_PURE_FE(n) || NODE_IS_FE_EFG(n))
#define NODE_IS_EFG(n)         (NODE_IS_PURE_EFG(n) || NODE_IS_FE_EFG(n))


typedef struct node_t {
  int           fe_id;
  int           efg_id;
  node_type_t   type;
  int           incident_to_fe_hexvol;
  int           incident_to_efg_hexvol;
  int           incident_to_iface_hexvol;
  WCRec         loc;
  int           constraint[3];
  double        prescr_displ[3];
  char         *symbolic_prescr_displ[3];
  GraphicObj   *g_rep;
  int           iface_id;
  int           in_curr_partition;
  int           on_ghost_only;
  double        d_m;
  int           surf_traction_tag;
} node_t;


typedef enum { FE_HEXVOL        = 0,
               EFG_HEXVOL       = 1,
               EFG_IFACE_HEXVOL = 2
             } hexvol_type_t;


#define HV_IS_FE(h) (h->type == FE_HEXVOL)
#define HV_IS_EFG(h) (h->type > 0)
#define HV_IS_PURE_EFG(h) (h->type == EFG_HEXVOL)
#define HV_IS_IFACE_EFG(h) (h->type == EFG_IFACE_HEXVOL)



typedef struct traction_components_t {
  double x, y, z;
}              traction_components_t;

  
typedef struct surface_traction_t {
  traction_components_t   node_traction[4];
  char                   *symbolic_traction[3];
}              surface_traction_t;


typedef struct hexvol_t {
  int                  fe_id;
  int                  efg_id;
  hexvol_type_t        type;
  hexvol_type_t        desired_type;
  node_t              *nodes[8];
  GraphicObj          *g_rep;
  int                  partition;
  int                  is_ghost;
  surface_traction_t  *face_traction[6];
  double               d_m; /* applies to nodes inside the volume */
  int                  quadrature_order;
  int                  split;
  int                  matid;
}              hexvol_t;



typedef struct face_load_t {
  int                     id;
  int                     type;
  double                  magnitude;
  ListOf(node_t *)        node_refs;
}              face_load_t;


typedef struct conc_load_t {
  int         id;
  node_t     *node;
  int         included_in_output;
  WCRec       components;
}              conc_load_t;


typedef struct output_request_t {
  int         id;
  node_t     *node;
}              output_request_t;


#define NODE_PICK_MASK (1L<<1)
#define HEX_PICK_MASK  (1L<<2)


#define new_instance_of_type(TYPE) NEWP_OBJ (TYPE)
#define realloc_array(P, N, T) NEWP_REALLOC_OBJ_ARRAY (P, N, T)
#define free_instance(PTR) NEWP_FREE_OBJ (PTR)


#include <assert.h>

#define Assert(EX)                                                            \
 ((EX)                                                                      \
  ? ((void)0)                                                                 \
  : (void)fprintf(stderr, "Assert failed:\n%s (%s, %d)\n",                    \
                  #EX, __FILE__, __LINE__))

#define DUPLICATE_STRING(dest, src)                                \
   if (((dest) = (char *)malloc((size_t)(strlen(src)+1))) == NULL) \
      fprintf(stderr, "malloc failed");                            \
   strcpy((dest), (src))



hexmesh_model_t *
new_hexmesh_model(char *description);

hexmesh_model_t *
destroy_hexmesh_model(hexmesh_model_t *fem);

node_t *
hexmesh_model_find_node(hexmesh_model_t *fem, WCRec *loc, double tol);

node_t *
hexmesh_model_add_node(hexmesh_model_t *fem, WCRec *loc, double tol);

hexvol_t *
hexmesh_model_add_hexvol(hexmesh_model_t *fem, node_t *nodes[8], 
                       GraphicObj *g_rep);

void
hexmesh_model_add_constraint(hexmesh_model_t *fem, int type,
                           ListOf(node_t *) nodes);


void
hexmesh_model_add_conc_load(hexmesh_model_t *fem, node_t *node, WCRec *components);

void
hexmesh_model_print(hexmesh_model_t *model, FILE *fp);

void
hexmesh_model_print_parallel(hexmesh_model_t *model, char *name);

void
hexmesh_model_print_TEES(hexmesh_model_t *model, FILE *fp,
                         int symbolic_tractions,
                         int nodes_at_hex_centers);

void
hexmesh_model_print_NDOT(hexmesh_model_t *model, FILE *fp);

void
hexmesh_model_set_properties(hexmesh_model_t *fem,
                        double thickness, double E, double nu);

void
hexmesh_model_add_output_request(hexmesh_model_t *fem, node_t *n);

hexvol_t *
hexmesh_model_find_hexvol(hexmesh_model_t *model, node_t *nodes[8]);

hexvol_t *
hexmesh_model_hexvol_by_g_rep(hexmesh_model_t *model, GraphicObj *g_rep);

int
hexmesh_get_hexvol_partition(hexmesh_model_t *model, hexvol_t *h);

void
hexmesh_set_hexvol_partition(hexmesh_model_t *model, hexvol_t *h, int partition);

int
hexmesh_set_function(int dir, char *expr);

void
hexmesh_model_add_surf_tractions(hexmesh_model_t *model,
                               LIST_OF(node_t *) nodes);

void
hexmesh_model_change_to_efg(hexmesh_model_t *model, hexvol_t *el);

void
hexmesh_set_model_quadrature_order(hexmesh_model_t *model, int qo);

void
hexmesh_compute_d_m(hexmesh_model_t *model, double max_d_m);

void
hexmesh_set_hexvol_quadrature_order(hexmesh_model_t *model, hexvol_t *h,
                                    int quadrature_order);

void
hexmesh_set_hexvol_subsplit(hexmesh_model_t *model, hexvol_t *h,
                                    int subsplit);

void
hexmesh_randomize_06_diagonal(hexmesh_model_t *model);

void
hexmesh_set_hexvol_mat_id (hexmesh_model_t *model, hexvol_t *h, int matid);

#endif
