#include "Esimple.h"
#include "hexmesh.h"



static void
reorder(WCRec *extrusion_vector, WCRec *qpts);


GraphicObj *
extrude_q_2_h(GraphicObj *q, WCRec *extrusion_vector)
{
  WCRec hpts[8];
  WCRec qpts[4];
  int npts;
  int i;

  if (!EGGetGraphicsGeometry(q, qpts, &npts))
    return NULL;

  reorder(extrusion_vector, qpts);

  for (i = 0; i < 4; i++) {
    hpts[i+4] = hpts[i] = qpts[i];
  }
  
  for (i = 4; i < 8; i++) {
    hpts[i].x += extrusion_vector->x;
    hpts[i].y += extrusion_vector->y;
    hpts[i].z += extrusion_vector->z;
  }

  return CreateHexahedron(hpts);
}


static void
reorder(WCRec *extrusion_vector, WCRec *qpts)
{
  WCRec normal, v0, v1;

  v0.x = qpts[1].x - qpts[0].x;
  v0.y = qpts[1].y - qpts[0].y;
  v0.z = qpts[1].z - qpts[0].z;
 
  v1.x = qpts[2].x - qpts[0].x;
  v1.y = qpts[2].y - qpts[0].y;
  v1.z = qpts[2].z - qpts[0].z;
  
  CrossProd3(&normal, &v0, &v1);
  if (DotProd3(&normal, extrusion_vector) < 0) { /* swap 1st and 3rd point */
    v0 = qpts[1];
    qpts[1] = qpts[3];
    qpts[3] = v0;
  }
}

GraphicObj *
sweep_q_2_h (GraphicObj *q, WCRec *center, WCRec *axis, double sweep_angle)
{
  WCRec hpts[8];
  WCRec qpts[4];
  int npts;
  int i;
  GraphicObj *qcopy;
  WCRec qc, dir={0,0,0}, axial;

  VECMAC_2_VEC (axial, =, (*axis));
  VECMAC_NORMALIZE (axial);
  VECMAC_SCALAR (axial, *=, sweep_angle);

  if (!EGGetGraphicsGeometry(q, qpts, &npts))
    return NULL;

  VECMAC_SCALAR (qc, =, 0);
  for (i = 0; i < 4; i++) {
    VECMAC_3_VEC (qc, +=, 0.25 * qpts[i], -, 4 * (*center));
  }
  
  VECMAC_CROSS (dir, =, axial, x, qc);
  
  for (i = 0; i < 4; i++) {
    hpts[i] = qpts[i];
  }

  qcopy = EGDeepCopyGraphics (q);

  EGRotateGraphic (qcopy, center, &axial);
  
  if (!EGGetGraphicsGeometry(qcopy, qpts, &npts))
    return NULL;

  for (i = 4; i < 8; i++) {
    hpts[i] = qpts[i-4];
  }

  return CreateHexahedron (hpts);
}
