#include "Esimple.h"
#include "hexmesh.h"
#include "qsplit.h"

#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

static int 
do_split(GraphicObj *h, int nx, int ny, int nz);


static LIST_OF(GraphicObj *) newquads = NULL;


LIST_OF(GraphicObj *)
split_q(GraphicObj *q, int nx, int ny, int nz)
{
  do_split(q, nx, ny, nz);
  return newquads;
}



static double
delta(WCRec qpts[4], int nx, int ny, int nz,
      int e1s, int e1e,
      int e2s, int e2e);




static int 
do_split(GraphicObj *q, int nx, int ny, int nz)
{
  WCRec qpts[4];
  WCRec pts[4];
  int npts;
  double dxi = 0, deta = 0;
  double xi = -1, eta = -1;
  double xiup = 1,  etaup = 1;
  int i;
  double n_mat[4];
#define LOC(pt, WHERE)                          \
  {                                             \
    G3D_q4_N(n_mat, WHERE);                     \
    pt.x = (  n_mat[0] * qpts[0].x              \
            + n_mat[1] * qpts[1].x              \
            + n_mat[2] * qpts[2].x              \
            + n_mat[3] * qpts[3].x);            \
    pt.y = (  n_mat[0] * qpts[0].y              \
            + n_mat[1] * qpts[1].y              \
            + n_mat[2] * qpts[2].y              \
            + n_mat[3] * qpts[3].y);            \
    pt.z = (  n_mat[0] * qpts[0].z              \
            + n_mat[1] * qpts[1].z              \
            + n_mat[2] * qpts[2].z              \
            + n_mat[3] * qpts[3].z);            \
  }
#define SHIFT(X, Y, SX, SY) X+SX, Y+SY
#define NQPTS(xi, eta, dxi, deta)               \
  {					       	\
    LOC(pts[0],  SHIFT(xi, eta, 0,     0));	\
    LOC(pts[1],  SHIFT(xi, eta, dxi,     0));	\
    LOC(pts[2],  SHIFT(xi, eta, dxi,  deta));	\
    LOC(pts[3],  SHIFT(xi, eta, 0,  deta));	\
  }

  if (newquads == NULL) newquads = make_list();
  flush_list(newquads, NOT_NODES);
  
  if (!EGGetGraphicsGeometry(q, qpts, &npts))
    return 0;

  nx = max(nx, 1);
  ny = max(ny, 1); 
  dxi    = delta(qpts, nx, ny, nz, 0, 1, 3, 2);
  deta   = delta(qpts, nx, ny, nz, 0, 3, 1, 2);

  xiup    = 1.0001 - dxi;
  etaup   = 1.0001 - deta;
  for (xi = -1; xi < xiup; xi += dxi) {
    for (eta = -1; eta < etaup; eta += deta) {
      NQPTS(xi, eta, dxi, deta);
      add_to_tail(newquads, CreateQuad3D(pts));
    }
  }

  return 1;
}



static double
delta(WCRec qpts[4], int nx, int ny, int nz, 
      int e1s, int e1e,
      int e2s, int e2e)
{
  WCRec v;
  double vn;
  double vpx, vpy, vpz;

  VECMAC_SCALAR(v, =, 0);
  VECMAC_3_VEC(v, +=, qpts[e1e], -, qpts[e1s]);
  VECMAC_3_VEC(v, +=, qpts[e2e], -, qpts[e2s]);
  vn = VECMAC_NORM(v);
  if (vn <= 0)
    return 2;

  VECMAC_NORMALIZE(v);

  if      (fabs(v.x) > fabs(v.y) &&
	   fabs(v.x) > fabs(v.z)
	   )
    return 2.0 / nx;
  else if (fabs(v.y) > fabs(v.x) &&
	   fabs(v.y) > fabs(v.z)
	   )
    return 2.0 / ny;
  else
    return 2.0 / nz;
}

