#include "Esimple.h"
#include "hexmesh.h"
#include "split.h"

#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

static int 
do_split(GraphicObj *h, int nx, int ny, int nz);


static LIST_OF(GraphicObj *) newhexes = NULL;


LIST_OF(GraphicObj *)
split_h(GraphicObj *h, int nx, int ny, int nz)
{
  do_split(h, nx, ny, nz);
  return newhexes;
}



static double
delta(WCRec hpts[8], int nx, int ny, int nz,
      int e1s, int e1e,
      int e2s, int e2e,
      int e3s, int e3e,
      int e4s, int e4e);




static int 
do_split(GraphicObj *h, int nx, int ny, int nz)
{
  WCRec hpts[8];
  WCRec pts[8];
  int npts;
  double dxi = 0, deta = 0, dtheta = 0;
  double xi = -1, eta = -1, theta = -1;
  double xiup = 1,  etaup = 1,  thetaup = 1;
  int i;
  double n_mat[8];
#define LOC(pt, WHERE)                        \
  {                                             \
    G3D_hex_N(n_mat, WHERE);                    \
    pt.x = (  n_mat[0] * hpts[0].x              \
            + n_mat[1] * hpts[1].x              \
            + n_mat[2] * hpts[2].x              \
            + n_mat[3] * hpts[3].x              \
            + n_mat[4] * hpts[4].x              \
            + n_mat[5] * hpts[5].x              \
            + n_mat[6] * hpts[6].x              \
            + n_mat[7] * hpts[7].x);            \
    pt.y = (  n_mat[0] * hpts[0].y              \
            + n_mat[1] * hpts[1].y              \
            + n_mat[2] * hpts[2].y              \
            + n_mat[3] * hpts[3].y              \
            + n_mat[4] * hpts[4].y              \
            + n_mat[5] * hpts[5].y              \
            + n_mat[6] * hpts[6].y              \
            + n_mat[7] * hpts[7].y);            \
    pt.z = (  n_mat[0] * hpts[0].z              \
            + n_mat[1] * hpts[1].z              \
            + n_mat[2] * hpts[2].z              \
            + n_mat[3] * hpts[3].z              \
            + n_mat[4] * hpts[4].z              \
            + n_mat[5] * hpts[5].z              \
            + n_mat[6] * hpts[6].z              \
            + n_mat[7] * hpts[7].z);            \
  }
#define SHIFT(X, Y, Z, SX, SY, SZ) X+SX, Y+SY, Z+SZ
#define NHPTS(xi, eta, theta, dxi, deta, dtheta) 		\
  {								\
    LOC(pts[0],  SHIFT(xi, eta, theta,   0,     0,       0));	\
    LOC(pts[1],  SHIFT(xi, eta, theta, dxi,     0,       0));	\
    LOC(pts[2],  SHIFT(xi, eta, theta, dxi,  deta,       0));	\
    LOC(pts[3],  SHIFT(xi, eta, theta,   0,  deta,       0));	\
    LOC(pts[4],  SHIFT(xi, eta, theta,   0,     0,  dtheta));	\
    LOC(pts[5],  SHIFT(xi, eta, theta, dxi,     0,  dtheta));	\
    LOC(pts[6],  SHIFT(xi, eta, theta, dxi,  deta,  dtheta));	\
    LOC(pts[7],  SHIFT(xi, eta, theta,   0,  deta,  dtheta));	\
  }

  if (newhexes == NULL) newhexes = make_list();
  flush_list(newhexes, NOT_NODES);
  
  if (!EGGetGraphicsGeometry(h, hpts, &npts))
    return 0;

  nx = max(nx, 1);
  ny = max(ny, 1); 
  nz = max(nz, 1);
  dxi    = delta(hpts, nx, ny, nz, 0, 1, 3, 2, 4, 5, 7, 6);
  deta   = delta(hpts, nx, ny, nz, 0, 3, 1, 2, 4, 7, 5, 6);
  dtheta = delta(hpts, nx, ny, nz, 0, 4, 1, 5, 2, 6, 3, 7);

  xiup    = 1.0001 - dxi;
  etaup   = 1.0001 - deta;
  thetaup = 1.0001 - dtheta;
  for (xi = -1; xi < xiup; xi += dxi) {
    for (eta = -1; eta < etaup; eta += deta) {
      for (theta = -1; theta < thetaup; theta += dtheta) {
	NHPTS(xi, eta, theta, dxi, deta, dtheta);
	add_to_tail(newhexes, CreateHexahedron(pts));
      }
    }
  }

  return 1;
}



static double
delta(WCRec hpts[8], int nx, int ny, int nz,
      int e1s, int e1e,
      int e2s, int e2e,
      int e3s, int e3e,
      int e4s, int e4e)
{
  WCRec v;
  double vn;
  double vpx, vpy, vpz;

  VECMAC_SCALAR(v, =, 0);
  VECMAC_3_VEC(v, +=, hpts[e1e], -, hpts[e1s]);
  VECMAC_3_VEC(v, +=, hpts[e2e], -, hpts[e2s]);
  VECMAC_3_VEC(v, +=, hpts[e3e], -, hpts[e3s]);
  VECMAC_3_VEC(v, +=, hpts[e4e], -, hpts[e4s]);
  vn = VECMAC_NORM(v);
  if (vn <= 0)
    return 2;

  VECMAC_NORMALIZE(v);

  if      (fabs(v.x) > fabs(v.y) &&
	   fabs(v.x) > fabs(v.z)
	   )
    return 2.0 / nx;
  else if (fabs(v.y) > fabs(v.x) &&
	   fabs(v.y) > fabs(v.z)
	   )
    return 2.0 / ny;
  else
    return 2.0 / nz;
}

