
#include "tfmap.h"


#define New(TYPE)             ((TYPE *) malloc(sizeof(TYPE)))
#define New_array(NELT, TYPE) ((TYPE *) calloc((NELT), (size_t) sizeof(TYPE)))
#define Realloc_array(PTR, NEW_SIZE, TYPE) \
                    ((TYPE *) realloc(PTR, (size_t)(NEW_SIZE * sizeof(TYPE))))
#define Free(P)               free(P)



/*
F(0,1)=E_1(1)=X_2(0)            X_2(xi)            F(1,1)=E_2(1)=X_2(1)


E_1(eta)                                                  E_2(eta)


F(0,0)=E_1(0)=X_1(0)            X_1(xi)            F(1,0)=E_2(0)=X_1(1)
*/

#define IPOL(C, par) bcurv_interpolate(par, C, &X, &Y)
#define INCR(mult) { *x += (mult) * X; *y += (mult) * Y; }


void
tfmap_interpolate(tfmap_t *tfm, double xi, double eta,
                  double *x, double *y)
{
  double X, Y;
#define X1 tfm->curve[0]
#define X2 tfm->curve[1]
#define E1 tfm->curve[2]
#define E2 tfm->curve[3]

  *x = *y = 0;

  IPOL(E1, eta); INCR(1 - xi);
  IPOL(E2, eta); INCR(xi);
  IPOL(X1, xi);  INCR(1 - eta);
  IPOL(X2, xi);  INCR(eta);

  IPOL(E2, 1);   INCR(- (xi * eta));           /* F(1,1) */
  IPOL(E1, 0);   INCR(- (1 - xi) * (1 - eta)); /* F(0,0) */
  IPOL(E2, 0);   INCR(- xi * (1 - eta));       /* F(1,0) */
  IPOL(E1, 1);   INCR(- (1 - xi) * eta);       /* F(0,1) */
#undef X1 
#undef X2
#undef E1
#undef E2   
}



tfmap_t *
tfmap_make_tfmap(b_curve_t X1,  /* curve xi, eta=0 */
                 b_curve_t X2,  /* curve xi, eta=1 */
                 b_curve_t E1,  /* curve xi=0, eta */
                 b_curve_t E2,  /* curve xi=1, eta */
                 double eps
                 )
{
  tfmap_t *tfm;
  double X, Y;
  double F00X, F00Y;
  double F10X, F10Y;
  double F11X, F11Y;
  double F01X, F01Y;
#define DO_GLUE(A, B)  A##B
#define GLUE(A, B)     DO_GLUE(A, B)
#define MOVE(F) { GLUE(F, X) = X; GLUE(F, Y) = Y; }
#define CHKLOC(F)                                       \
  {                                                     \
    if (fabs(GLUE(F, X) - X) > eps) return NULL;        \
    if (fabs(GLUE(F, Y) - Y) > eps) return NULL;        \
  }
  
  IPOL(E2, 1);   MOVE(F11);
  IPOL(E1, 0);   MOVE(F00);
  IPOL(E2, 0);   MOVE(F10);
  IPOL(E1, 1);   MOVE(F01);

  IPOL(X2, 1); CHKLOC(F11);
  IPOL(X1, 0); CHKLOC(F00);
  IPOL(X1, 1); CHKLOC(F10);
  IPOL(X2, 0); CHKLOC(F01);

  tfm = New(tfmap_t);
  if (tfm == NULL) return NULL;

  tfm->curve[0] = X1;
  tfm->curve[1] = X2;
  tfm->curve[2] = E1;
  tfm->curve[3] = E2;

  return tfm;
}

